mod chrony;

use clap::Parser;
use env_logger;
use futures::{future::join_all, prelude::*};
use log::{error, info, warn};
use std::{path::Path, time::Duration};
use std::net::SocketAddr;
use tokio::runtime::Handle;

use crate::chrony::*;
use chimemon::*;

const PROGRAM_NAME: &str = "chimemon";
const VERSION: &str = "0.0.1";

const DEFAULT_RETRY_DELAY: Duration = Duration::new(60, 0);
const MAX_RETRY_DELAY: Duration = Duration::new(1800, 0);

#[derive(Parser)]
struct Args {
    /// TOML configuration to load
    #[arg(short, long, default_value_t = String::from("config.toml"))]
    config_file: String,
}

async fn do_chrony_poll(
    config: &Config,
    chrony: &ChronyClient,
    influx: &influxdb2::Client,
) -> Result<(), Box<dyn std::error::Error>> {
    let tracking = chrony.get_tracking().await?;
    let sources = chrony.get_sources().await?;

    let mut frame = Vec::new();

    let tracking_data = datapoint_from_tracking(&tracking, &config)?;

    info!(target: "chrony", "Writing tracking data: {:?}", tracking_data);

    frame.push(tracking_data);
    for ds in sources {
        let source_data = datapoint_from_sourcedata(&ds, &config)?;
        info!(target: "chrony", "Writing source data: {:?}", source_data);
        frame.push(source_data);
    }

    influx
        .write(&config.influxdb.bucket, stream::iter(frame))
        .await?;
    Ok(())
}

#[tokio::main(flavor = "current_thread")]
async fn main() -> Result<(), Box<dyn std::error::Error>> {
    env_logger::init();
    let args = Args::parse();
    info!("{} v{} starting...", PROGRAM_NAME, VERSION);
    let config = load_config(Path::new(&args.config_file))?;
    let mut tasks = Vec::new();

    info!(
        "Connecting to influxdb {} org: {} using token",
        &config.influxdb.url, &config.influxdb.org
    );
    let influx = influxdb2::Client::new(
        &config.influxdb.url,
        &config.influxdb.org,
        &config.influxdb.token,
    );

    if config.sources.chrony.enabled {
        let chrony = ChronyClient::new(
            chrony_candm::Client::spawn(&Handle::current(), Default::default()),
            SocketAddr::new(
                config.sources.chrony.host,
                config.sources.chrony.port,
            ),
            Duration::from_secs(config.sources.chrony.timeout),
        );

        let mut retry_delay = DEFAULT_RETRY_DELAY;
        let mut fail_count = 0;

        let interval = Duration::new(config.sources.chrony.poll_interval, 0);
        let mut interval = tokio::time::interval(interval);
        interval.set_missed_tick_behavior(tokio::time::MissedTickBehavior::Skip);

        tasks.push(tokio::spawn(async move {
            info!(target: "chrony", "Chrony task started");
            loop {
                interval.tick().await;
                match do_chrony_poll(&config, &chrony, &influx.to_owned()).await {
                    Ok(_) => {
                        fail_count = 0;
                        retry_delay = DEFAULT_RETRY_DELAY;
                    }
                    Err(e) => {
                        warn!(target: "chrony", "Error in chrony task: {}", e.to_string());
                        if fail_count % 4 == 0 {
                            // exponential backoff, every 4 failures double the retry timer, cap at max
                            retry_delay = std::cmp::min(MAX_RETRY_DELAY, retry_delay * 2);
                        }
                        fail_count += 1;
                    }
                }
                if fail_count > 0 {
                    tokio::time::sleep(retry_delay).await;
                    interval.reset();
                };
            }
        }));
    }

    join_all(tasks).await;

    Ok(())
}
